/**
 * Copyright Marcus Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of LibDesQ (https://gitlab.com/DesQ/libdesq)
 * This library contains various  Core and Gui classes which are used
 * through out the DesQ Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include "desq/Settings.hpp"
#include "desq/Utils.hpp"
#include "desq-config.h"
#include <QtDBus>

static inline int getFormFactor() {
    /*
     *
     * Form Factors
     * ------------
     *
     * 0: Desktop - Typically 14" to 16" screens
     * 1: Tablets - Typically  7" to 10" screens
     * 2: Mobiles - Typically  4" to  6" screens
     * 3: Monitor - Typically larger than 16" - huge sizes, we're not built for these screens.
     *
     */

    QSizeF screenSize = QGuiApplication::primaryScreen()->physicalSize();
    double diag       = sqrt( pow( screenSize.width(), 2 ) + pow( screenSize.height(), 2 ) ) / 25.4;

    //- Deady large screen mobiles
    if ( diag <= 6.5 ) {
        return 2;
    }

    //- Tablets and small netbooks
    else if ( diag <= 10.1 ) {
        return 1;
    }

    //- Typical desktops
    else if ( diag <= 16.0 ) {
        return 0;
    }

    // Large displays
    else {
        return 3;
    }

    /* Default value */
    return 0;
}


DesQ::Settings::Settings( QString app ) {
    system = new QSettings( QString( ConfigPath ) + "DesQ" + app + ".conf", QSettings::IniFormat );
    user   = new QSettings( "DesQ", app );

    QFileSystemWatcher *fsw = new QFileSystemWatcher( this );

    fsw->addPath( system->fileName() );
    fsw->addPath( user->fileName() );
    connect(
        fsw, &QFileSystemWatcher::fileChanged, [ = ]( const QString& file ) {
            /* User config */
            if ( file.startsWith( "/home" ) ) {
                user->sync();
                emitSettings( user );
            }

            else {
                system->sync();
                emitSettings( system );
            }

            /* File deleted and created again: add it to the watcher */
            if ( not fsw->files().contains( file ) and QFile::exists( file ) ) {
                fsw->addPath( file );
            }
        }
    );
}


DesQ::Settings::DesQProxy DesQ::Settings::value( const QString& key ) {
    QVariant rawVal( rawValue( key ) );

    if ( key == "Font" or key.contains( "FontFamily" ) ) {
        if ( (QMetaType::Type)rawVal.type() == QMetaType::QString ) {
            if ( (rawVal.toString() == "default") or (rawVal.toString() == "sans-serif") ) {
                rawVal = QFontDatabase::systemFont( QFontDatabase::GeneralFont );
            }

            else if ( (rawVal.toString() == "fixed") or (rawVal.toString() == "monospace") ) {
                rawVal = QFontDatabase::systemFont( QFontDatabase::FixedFont );
            }
        }
    }

    return DesQProxy{ rawVal };
}


void DesQ::Settings::setValue( const QString& key, QVariant value ) {
    if ( not system->contains( key ) and not key.startsWith( "Session/" ) ) {
        qDebug() << "Unknown key:" << key;
        qDebug() << "Report this incident to the developers.";

        return;
    }

    user->setValue( key, value );
    user->sync();
}


QVariant DesQ::Settings::rawValue( const QString& key ) {
    /*
     * SafetyNet: if the @system does not contain this key, then the setting is not available
     * However, session data will not be available in @system, so ignore the 'Session/' stuff
     */
    if ( not system->contains( key ) and not key.startsWith( "Session/" ) ) {
        return QVariant();
    }

    if ( key == "FormFactor" ) {
        return getFormFactor();
    }

    /**
     * If there is not key for a session return an empty QVariant.
     * It's upto the the user to handle it the way they wish.
     */
    if ( key.startsWith( "Session/" ) and (not user->contains( key ) ) ) {
        return QVariant();
    }

    QVariant value = (user->contains( key ) ? user->value( key ) : system->value( key ) );

    /* TouchMode: As seen by the oser */
    if ( key == "TouchMode" ) {
        int touch = value.toInt();
        switch ( touch ) {
            /* Auto detect */
            case 0: {
                QProcess proc;
                proc.start( "udevadm", QStringList() << "info" << "--export-db" );
                proc.waitForFinished( -1 );

                QString output = QString::fromLocal8Bit( proc.readAllStandardOutput() + '\n' + proc.readAllStandardError() );

                if ( output.contains( "ID_INPUT_TOUCHSCREEN=1" ) ) {
                    return QVariant( true );
                }

                return QVariant( false );
            }

            /* TouchMode is on */
            case 1: {
                return QVariant( true );
            }

            /* TouchMode is off */
            case 2: {
                return QVariant( false );
            }

            default: {
                return QVariant( false );
            }
        }
    }

    /* TouchModeRaw: Raw value saved in the file */
    else if ( key == "TouchModeRaw" ) {
        return value;
    }

    /* Get the default value for the key */
    else {
        if ( (QMetaType::Type)value.type() == QMetaType::QString ) {
            QString strValue = value.toString();

            /** Replace ~ with home path, but only if the string begins with ~ */
            if ( strValue.startsWith( "~/" ) or strValue == "~" ) {
                strValue = strValue.replace( "~", QDir::homePath() );
            }

            /** Envisonment variable */
            else if ( strValue.startsWith( "$" ) ) {
                QString envVar      = strValue.split( "/" ).takeFirst();
                QString envVarValue = qEnvironmentVariable( envVar.toUtf8().constData() );

                if ( not envVarValue.isEmpty() ) {
                    strValue = strValue.replace( envVar, envVarValue );
                }
            }

            return QVariant( strValue );
        }
    }

    return value;
}


void DesQ::Settings::emitSettings( QSettings *sett ) {
    /* If the global settings changed */
    if ( sett == system ) {
        qDebug() << "Global settings have changed. This generally does not happen. " \
            "If your app was upgraded, then it's strongly recommended that "         \
            "you restart your app. Otherwise you may face compatibility issues.";

        return;
    }

    /* We are bothered only about changes to system settings */
    SettingsMap newMap;

    for ( QString key: user->allKeys() ) {
        /* Store the (key, vaule) pair */
        newMap[ key ] = user->value( key );

        /* Not important for a running app, is it? */
        if ( key.startsWith( "Session/" ) ) {
            continue;
        }

        /* If the key isn't in the system config, don't alert the user */
        if ( not system->contains( key ) ) {
            continue;
        }

        /* If the value is same as the old one, nothing to do */
        if ( userMap[ key ] == user->value( key ) ) {
            continue;
        }

        /* New key, or value has changed */
        emit settingChanged( key, newMap[ key ] );
    }

    /*
     * A key may have been removed from the config file.
     * Check and return system config value
     */
    for ( QString key: userMap.keys() ) {
        if ( not newMap.contains( key ) ) {
            if ( system->contains( key ) ) {
                emit settingChanged( key, system->value( key ) );
            }
        }
    }

    /* Clear the old map, and store the new map in it. */
    userMap.clear();
    userMap = newMap;
}
